"""scrapli.channel.base_channel"""
import re
from abc import ABC
from logging import getLogger
from typing import List, Optional, Tuple, Union

from scrapli.helper import get_prompt_pattern, normalize_lines
from scrapli.transport.async_transport import AsyncTransport
from scrapli.transport.transport import Transport

CHANNEL_ARGS = (
    "transport",
    "comms_prompt_pattern",
    "comms_return_char",
    "comms_ansi",
    "timeout_ops",
)


class ChannelBase(ABC):
    def __init__(
        self,
        transport: Union[Transport, AsyncTransport],
        comms_prompt_pattern: str = r"^[a-z0-9.\-@()/:]{1,32}[#>$]$",
        comms_return_char: str = "\n",
        comms_ansi: bool = False,
        comms_auto_expand: bool = False,
        timeout_ops: float = 10,
    ):
        """
        Channel Object

        Args:
            transport: Transport object of any transport provider (system|telnet or a plugin)
                transport could in theory be any transport as long as it provides a read and a write
                method... obviously its probably always going to be scrapli transport though
            comms_prompt_pattern: raw string regex pattern -- use `^` and `$` for multi-line!
            comms_return_char: character to use to send returns to host
            comms_ansi: True/False strip comms_ansi characters from output
            comms_auto_expand: bool to indicate if a device auto-expands commands, for example
                juniper devices without `cli complete-on-space` disabled will convert `config` to
                `configuration` after entering a space character after `config`; because scrapli
                reads the channel until each command is entered, the command changing from `config`
                to `configuration` will cause scrapli (by default) to never think the command has
                been entered. Setting this value to `True` will force scrapli to zip the split lists
                of inputs and outputs together to determine if each read output starts with the
                corresponding input. For example, if the inputs are "sho ver" and the read output is
                "show version", scrapli will zip the split strings together and confirm that in fact
                "show" starts with "sho" and "version" starts with "ver", confirming that the
                commands that were input were input properly. This is disabled by default, as it is
                preferable to disable this type of behavior via the device itself if possible.
            timeout_ops: timeout in seconds for channel operations (reads/writes)

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.logger = getLogger(f"scrapli.channel-{transport.host}")

        self.transport = transport
        self.comms_prompt_pattern = comms_prompt_pattern
        self.comms_return_char = comms_return_char
        self.comms_ansi = comms_ansi
        self.comms_auto_expand = comms_auto_expand
        self.timeout_ops = timeout_ops

    def __str__(self) -> str:
        """
        Magic str method for Channel

        Args:
            N/A

        Returns:
            str: str for class object

        Raises:
            N/A

        """
        return "scrapli Channel Object"

    def __repr__(self) -> str:
        """
        Magic repr method for Channel

        Args:
            N/A

        Returns:
            str: repr for class object

        Raises:
            N/A

        """
        class_dict = self.__dict__.copy()
        class_dict.pop("transport")
        class_dict["logger"] = self.logger.name
        return f"scrapli Channel {class_dict}"

    def _restructure_output(self, output: bytes, strip_prompt: bool = False) -> bytes:
        """
        Clean up preceding empty lines, and strip prompt if desired

        Args:
            output: bytes from channel
            strip_prompt: bool True/False whether to strip prompt or not

        Returns:
            bytes: output of joined output lines optionally with prompt removed

        Raises:
            N/A

        """
        output = normalize_lines(output=output)

        if strip_prompt:
            # could be compiled elsewhere, but allow users to modify the prompt whenever they want
            prompt_pattern = get_prompt_pattern(prompt="", class_prompt=self.comms_prompt_pattern)
            output = re.sub(pattern=prompt_pattern, repl=b"", string=output)

        # lstrip the return character out of the final result before storing, also remove any extra
        # whitespace to the right if any
        output = output.lstrip(self.comms_return_char.encode()).rstrip()
        return output

    @staticmethod
    def _process_auto_expand(output: bytes, channel_input: bytes) -> bool:
        """
        Determine if output has been auto expanded to canonical syntax

        Args:
            output: output from the device
            channel_input: command input to the device

        Returns:
            bool: True if it appears the output was auto-expanded, otherwise False

        Raises:
            N/A

        """
        channel_input_split = channel_input.split()
        return all(
            _channel_output.startswith(_channel_input)
            for _channel_input, _channel_output in zip(channel_input_split, output.split())
        )

    def _send_return(self) -> None:
        """
        Send return char to device

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.transport.write(channel_input=self.comms_return_char)
        self.logger.debug(f"Write (sending return character): {repr(self.comms_return_char)}")

    @staticmethod
    def _pre_send_input(channel_input: str) -> None:
        """
        Handle pre "send_input" tasks for consistency between sync/async versions

        Args:
            channel_input: string input to send to channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            TypeError: if input is anything but a string

        """
        if not isinstance(channel_input, str):
            raise TypeError(f"`send_input` expects a single string, got {type(channel_input)}.")

    @staticmethod
    def _pre_send_inputs_interact(interact_events: List[Tuple[str, str, Optional[bool]]]) -> None:
        """
        Handle pre "send_inputs_interact" tasks for consistency between sync/async versions

        Args:
            interact_events: interact events passed to `send_inputs_interact`

        Returns:
            N/A  # noqa: DAR202

        Raises:
            TypeError: if input is anything but a string

        """
        if not isinstance(interact_events, list):
            raise TypeError(f"`interact_events` expects a List, got {type(interact_events)}")

    def _post_send_inputs_interact(self, output: bytes) -> Tuple[bytes, bytes]:
        """
        Handle pre "send_inputs_interact" tasks for consistency between sync/async versions

        Args:
            output: output from `send_inputs_interact` method

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        processed_output = self._restructure_output(output=output, strip_prompt=False)
        raw_result = output
        processed_result = processed_output
        return raw_result, processed_result



import asyncio
import time
from threading import Lock

import pytest

from scrapli.decorators import OperationTimeout, TimeoutModifier, requires_open_session
from scrapli.driver import AsyncScrape, Scrape
from scrapli.driver.base_driver import ScrapeBase
from scrapli.exceptions import ConnectionNotOpened, ScrapliTimeout


class SlowClass(ScrapeBase):
    def __init__(self):
        # subclass base scrape to have a logger setup and such
        # set transport to telnet as it works on all platforms and is standard library!
        super().__init__(host="localhost", transport="telnet")
        self.timeout_test = 0.25
        self.timeout_exit = True
        self.session_lock = Lock()
        self.session_lock.acquire()

    @OperationTimeout("timeout_test")
    def slow_function(self):
        time.sleep(1)

    @OperationTimeout("timeout_test")
    def fast_function(self):
        return "fast"

    @OperationTimeout("non_existent_class_attr")
    def confused_function(self):
        return "fast"

    @OperationTimeout("timeout_test")
    async def async_slow_function(self):
        await asyncio.sleep(1)

    @OperationTimeout("timeout_test")
    async def async_fast_function(self):
        return "fast"

    @OperationTimeout("non_existent_class_attr")
    async def async_confused_function(self):
        return "fast"

    def close(self):
        return

    @requires_open_session()
    def raise_attribute_error(self):
        raise AttributeError


def test_operation_timeout_timeout():
    slow = SlowClass()
    with pytest.raises(ScrapliTimeout):
        slow.slow_function()


def test_operation_timeout_success():
    slow = SlowClass()
    result = slow.fast_function()
    assert result == "fast"


def test_operation_timeout_no_class_attr():
    slow = SlowClass()
    result = slow.confused_function()
    assert result == "fast"


@pytest.mark.asyncio
async def test_async_operation_timeout_timeout():
    slow = SlowClass()
    with pytest.raises(ScrapliTimeout):
        await slow.async_slow_function()


@pytest.mark.asyncio
async def test_async_operation_timeout_success():
    slow = SlowClass()
    result = await slow.async_fast_function()
    assert result == "fast"


@pytest.mark.asyncio
async def test_async_operation_timeout_no_class_attr():
    slow = SlowClass()
    result = await slow.async_confused_function()
    assert result == "fast"


def test_requires_open_session():
    slow = SlowClass()
    with pytest.raises(ConnectionNotOpened) as exc:
        slow.raise_attribute_error()
    assert (
        str(exc.value)
        == "Attempting to call method that requires an open connection, but connection is not open. "
        "Call the `.open()` method of your connection object, or use a context manager to ensue "
        "your connection has been opened."
    )


@pytest.mark.parametrize(
    "timeout_ops",
    [
        999,
        30,
        None,
    ],
    ids=["timeout_modified", "timeout_unchanged", "timeout_not_provided"],
)
def test_timeout_modifier(monkeypatch, timeout_ops):
    scrape = Scrape(host="localhost")
    assert scrape.timeout_ops == 30

    @TimeoutModifier()
    def test_timeout_modifier(cls, timeout_ops):
        return cls.timeout_ops

    # stupid patch but does confirm the timeout modifier works as expected!
    monkeypatch.setattr(Scrape, "open", test_timeout_modifier)
    modified_timeout = scrape.open(timeout_ops=timeout_ops)
    assert modified_timeout == timeout_ops if timeout_ops else 30
    assert scrape.timeout_ops == 30


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "timeout_ops",
    [
        999,
        30,
        None,
    ],
    ids=["timeout_modified", "timeout_unchanged", "timeout_not_provided"],
)
async def test_timeout_modifier_async(monkeypatch, timeout_ops):
    scrape = AsyncScrape(host="localhost", transport="asyncssh")
    assert scrape.timeout_ops == 30

    @TimeoutModifier()
    async def test_timeout_modifier(cls, timeout_ops):
        return cls.timeout_ops

    # stupid patch but does confirm the timeout modifier works as expected!
    monkeypatch.setattr(AsyncScrape, "open", test_timeout_modifier)
    modified_timeout = await scrape.open(timeout_ops=timeout_ops)
    assert modified_timeout == timeout_ops if timeout_ops else 30
    assert scrape.timeout_ops == 30



"""scrapli.transport.async_transport"""
from abc import ABC, abstractmethod

from scrapli.transport.base_transport import TransportBase


class AsyncTransport(TransportBase, ABC):
    @abstractmethod
    async def open(self) -> None:  # pylint: disable=W0236
        """
        Async open channel, acquire pty, request interactive shell

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    async def read(self) -> bytes:  # pylint: disable=W0236
        """
        Async read data from the channel

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def write(self, channel_input: str) -> None:
        """
        Write data to the channel

        Args:
            channel_input: string to send to channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """



"""scrapli.channel"""
from scrapli.channel.async_channel import AsyncChannel
from scrapli.channel.base_channel import CHANNEL_ARGS
from scrapli.channel.channel import Channel

__all__ = ("AsyncChannel", "Channel", "CHANNEL_ARGS")



import pytest

from scrapli.driver.core import IOSXEDriver
from scrapli.exceptions import TransportPluginError

from ...test_data.devices import DEVICES


def test_context_manager(sync_cisco_iosxe_conn):
    with sync_cisco_iosxe_conn as conn:
        assert conn.isalive() is True
    assert conn.isalive() is False


def test_transport_bool(sync_cisco_iosxe_conn):
    sync_cisco_iosxe_conn.open()
    assert bool(sync_cisco_iosxe_conn.transport) is True
    sync_cisco_iosxe_conn.close()
    assert bool(sync_cisco_iosxe_conn.transport) is False


def test_transport_encrypted_private_key(sync_cisco_iosxe_conn_encrypted_key):
    sync_cisco_iosxe_conn_encrypted_key.open()
    assert bool(sync_cisco_iosxe_conn_encrypted_key.transport) is True
    sync_cisco_iosxe_conn_encrypted_key.close()
    assert bool(sync_cisco_iosxe_conn_encrypted_key.transport) is False


def test_non_sync_transport():
    device = DEVICES["mock_cisco_iosxe"].copy()
    with pytest.raises(TransportPluginError) as exc:
        IOSXEDriver(transport="asyncssh", **device)
    assert (
        str(exc.value)
        == "Attempting to use transport type asyncssh with a sync driver, must use a non-asyncio transport"
    )


def test_timeout_ops_property(sync_cisco_iosxe_conn):
    # note that there is "sync" and "async" version of this test as the property setter/getter lives
    # in the sync and async drivers -- didnt want to deal w/ having the property in the base driver
    # and having to deal w/ _get_timeout_ops/_set_timeout_ops methods to get overridden by the sync
    # and async drivers
    assert sync_cisco_iosxe_conn.timeout_ops == 30
    assert sync_cisco_iosxe_conn.channel.timeout_ops == 30
    sync_cisco_iosxe_conn.timeout_ops = 5
    assert sync_cisco_iosxe_conn.timeout_ops == 5
    assert sync_cisco_iosxe_conn.channel.timeout_ops == 5



"""scrapli.transport.socket"""
import socket
from logging import getLogger
from typing import Optional

from scrapli.exceptions import ScrapliTimeout


class Socket:
    def __init__(self, host: str, port: int, timeout: int):
        """
        Socket object

        Args:
            host: host to connect to
            port: port to connect to
            timeout: timeout in seconds

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.logger = getLogger(f"scrapli.socket-{host}")
        self.host: str = host
        self.port: int = port
        self.timeout: int = timeout
        self.sock: Optional[socket.socket] = None

    def __bool__(self) -> bool:
        """
        Magic bool method for Socket

        Args:
            N/A

        Returns:
            bool: True/False if socket is alive or not

        Raises:
            N/A

        """
        return self.socket_isalive()

    def __str__(self) -> str:
        """
        Magic str method for Socket

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        return f"Socket Object for host {self.host}"

    def __repr__(self) -> str:
        """
        Magic repr method for Socket

        Args:
            N/A

        Returns:
            str: repr for class object

        Raises:
            N/A

        """
        class_dict = self.__dict__.copy()
        class_dict["logger"] = self.logger.name
        return f"Socket {class_dict}"

    def socket_open(self) -> None:
        """
        Open underlying socket

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            ConnectionRefusedError: if socket refuses connection
            ScrapliTimeout: if socket connection times out

        """
        if not self.socket_isalive():
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.settimeout(self.timeout)
            try:
                self.sock.connect((self.host, self.port))
            except ConnectionRefusedError as exc:
                self.logger.critical(
                    f"Connection refused trying to open socket to {self.host} on port {self.port}"
                )
                raise ConnectionRefusedError(
                    f"Connection refused trying to open socket to {self.host} on port {self.port}"
                ) from exc
            except socket.timeout as exc:
                self.logger.critical(
                    f"Timed out trying to open socket to {self.host} on port {self.port}"
                )
                raise ScrapliTimeout(
                    f"Timed out trying to open socket to {self.host} on port {self.port}"
                ) from exc
            self.logger.debug(f"Socket to host {self.host} opened")

    def socket_close(self) -> None:
        """
        Close socket

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.socket_isalive() and isinstance(self.sock, socket.socket):
            self.sock.close()
            self.logger.debug(f"Socket to host {self.host} closed")

    def socket_isalive(self) -> bool:
        """
        Check if socket is alive

        Args:
            N/A

        Returns:
            bool True/False if socket is alive

        Raises:
            N/A

        """
        try:
            if isinstance(self.sock, socket.socket):
                self.sock.send(b"")
                return True
        except OSError:
            self.logger.debug(f"Socket to host {self.host} is not alive")
            return False
        return False



from pathlib import Path

import pytest

import scrapli
from scrapli.channel import Channel
from scrapli.driver.core import IOSXEDriver
from scrapli.exceptions import ScrapliTimeout
from scrapli.transport import Transport

from ...test_data.unit_test_cases import TEST_CASES

TEST_DATA_DIR = f"{Path(scrapli.__file__).parents[1]}/tests/test_data"


def test_check_kwargs_comms_prompt_pattern():
    with pytest.warns(UserWarning) as warn:
        conn = IOSXEDriver(host="localhost", comms_prompt_pattern="something")
    assert (
        conn.comms_prompt_pattern
        == "(^[a-z0-9.\\-_@()/:]{1,63}>$)|(^[a-z0-9.\\-_@/:]{1,63}#$)|(^[a-z0-9.\\-_@/:]{1,63}\\(conf[a-z0-9.\\-@/:\\+]{"
        "0,32}\\)#$)"
    )
    assert (
        str(warn[0].message) == "\n***** `comms_prompt_pattern` found in kwargs! "
        "*****************************************\n`comms_prompt_pattern` is ignored (dropped) when using network "
        "drivers. If you wish to modify the patterns for any network driver sub-classes, please do so by modifying "
        "or providing your own `privilege_levels`.\n***** `comms_prompt_pattern` found in kwargs! "
        "*****************************************"
    )


def test_init(sync_cisco_iosxe_conn):
    """Test that all arguments get properly passed from driver/transport to channel on init"""
    assert isinstance(sync_cisco_iosxe_conn.channel.transport, Transport)
    assert isinstance(sync_cisco_iosxe_conn.channel, Channel)
    assert sync_cisco_iosxe_conn.auth_secondary == "scrapli"
    assert list(sync_cisco_iosxe_conn.privilege_levels.keys()) == [
        "exec",
        "privilege_exec",
        "configuration",
    ]
    assert sync_cisco_iosxe_conn.default_desired_privilege_level == "privilege_exec"
    assert sync_cisco_iosxe_conn.textfsm_platform == "cisco_iosxe"
    assert sync_cisco_iosxe_conn.genie_platform == "iosxe"
    assert sync_cisco_iosxe_conn.failed_when_contains == [
        "% Ambiguous command",
        "% Incomplete command",
        "% Invalid input detected",
        "% Unknown command",
    ]


def test_escalate(sync_cisco_iosxe_conn):
    expected_prompt_privilege_exec = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["privilege_exec"]
    expected_prompt_configuration = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["configuration"]
    sync_cisco_iosxe_conn.open()
    assert sync_cisco_iosxe_conn._current_priv_level.name == "privilege_exec"
    assert sync_cisco_iosxe_conn.get_prompt() == expected_prompt_privilege_exec
    sync_cisco_iosxe_conn._escalate(
        escalate_priv=sync_cisco_iosxe_conn.privilege_levels["configuration"]
    )
    assert sync_cisco_iosxe_conn.get_prompt() == expected_prompt_configuration


def test_deescalate(sync_cisco_iosxe_conn):
    expected_prompt_privilege_exec = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["privilege_exec"]
    expected_prompt_exec = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["exec"]
    sync_cisco_iosxe_conn.open()
    assert sync_cisco_iosxe_conn._current_priv_level.name == "privilege_exec"
    assert sync_cisco_iosxe_conn.get_prompt() == expected_prompt_privilege_exec
    sync_cisco_iosxe_conn._deescalate(
        current_priv=sync_cisco_iosxe_conn.privilege_levels["privilege_exec"]
    )
    assert sync_cisco_iosxe_conn.get_prompt() == expected_prompt_exec


def test_acquire_priv(sync_cisco_iosxe_conn):
    expected_prompt_privilege_exec = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["privilege_exec"]
    expected_prompt_exec = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["exec"]
    expected_prompt_configuration = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["configuration"]
    sync_cisco_iosxe_conn.open()
    assert sync_cisco_iosxe_conn._current_priv_level.name == "privilege_exec"
    assert sync_cisco_iosxe_conn.get_prompt() == expected_prompt_privilege_exec
    sync_cisco_iosxe_conn.acquire_priv(desired_priv="exec")
    assert sync_cisco_iosxe_conn._current_priv_level.name == "exec"
    assert sync_cisco_iosxe_conn.get_prompt() == expected_prompt_exec
    sync_cisco_iosxe_conn.acquire_priv(desired_priv="configuration")
    assert sync_cisco_iosxe_conn._current_priv_level.name == "configuration"
    assert sync_cisco_iosxe_conn.get_prompt() == expected_prompt_configuration


def test_send_command_not_desired_priv_level(sync_cisco_iosxe_conn):
    expected_prompt_privilege_exec = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["privilege_exec"]
    sync_cisco_iosxe_conn.open()
    sync_cisco_iosxe_conn.send_command(command="terminal length 0")
    sync_cisco_iosxe_conn.acquire_priv(desired_priv="exec")
    sync_cisco_iosxe_conn.send_command(command="show version")
    _ = sync_cisco_iosxe_conn.send_command(command="show version")
    assert sync_cisco_iosxe_conn._current_priv_level.name == "privilege_exec"
    assert sync_cisco_iosxe_conn.get_prompt() == expected_prompt_privilege_exec


@pytest.mark.parametrize(
    "strip_prompt",
    [True, False],
    ids=["strip_prompt", "no_strip_prompt"],
)
def test_send_command(sync_cisco_iosxe_conn, strip_prompt):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_input"]["raw_result"]
    expected_processed = (
        TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["strip"]
        if strip_prompt
        else TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["no_strip"]
    )
    sync_cisco_iosxe_conn.open()
    sync_cisco_iosxe_conn.send_command(command="terminal length 0")
    response = sync_cisco_iosxe_conn.send_command(command="show version", strip_prompt=strip_prompt)
    assert response.raw_result == expected_raw.encode()
    assert response.result == expected_processed


@pytest.mark.parametrize(
    "strip_prompt",
    [True, False],
    ids=["strip_prompt", "no_strip_prompt"],
)
def test_send_commands(sync_cisco_iosxe_conn, strip_prompt):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_input"]["raw_result"]
    expected_processed = (
        TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["strip"]
        if strip_prompt
        else TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["no_strip"]
    )
    sync_cisco_iosxe_conn.open()
    multi_response = sync_cisco_iosxe_conn.send_commands(
        commands=["terminal length 0", "show version"], strip_prompt=strip_prompt
    )
    assert multi_response[1].raw_result == expected_raw.encode()
    assert multi_response[1].result == expected_processed


@pytest.mark.parametrize(
    "strip_prompt",
    [True, False],
    ids=["strip_prompt", "no_strip_prompt"],
)
def test_send_commands_from_file(sync_cisco_iosxe_conn, strip_prompt):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_input"]["raw_result"]
    expected_processed = (
        TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["strip"]
        if strip_prompt
        else TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["no_strip"]
    )
    sync_cisco_iosxe_conn.open()
    sync_cisco_iosxe_conn.send_command(command="terminal length 0")
    multi_response = sync_cisco_iosxe_conn.send_commands_from_file(
        file=f"{TEST_DATA_DIR}/files/cisco_iosxe_commands", strip_prompt=strip_prompt
    )
    assert multi_response[0].raw_result == expected_raw.encode()
    assert multi_response[0].result == expected_processed


def test_send_interactive(sync_cisco_iosxe_conn):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["raw_result"]
    expected_processed = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["processed_result"]
    interact_events = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["interact_events"]
    sync_cisco_iosxe_conn.open()
    response = sync_cisco_iosxe_conn.send_interactive(interact_events=interact_events)
    assert expected_raw.encode() in response.raw_result
    assert expected_processed in response.result


def test_get_prompt(sync_cisco_iosxe_conn):
    expected_prompt = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["privilege_exec"]
    sync_cisco_iosxe_conn.open()
    found_prompt = sync_cisco_iosxe_conn.get_prompt()
    assert found_prompt == expected_prompt


@pytest.mark.parametrize(
    "auth_secondary",
    [("scrapli", True), ("", False)],
    ids=["auth_secondary_provided", "auth_secondary_missed"],
)
def test_auth_required_no_auth_secondary(sync_cisco_iosxe_conn, auth_secondary):
    sync_cisco_iosxe_conn.auth_secondary = auth_secondary[0]
    sync_cisco_iosxe_conn.open()
    sync_cisco_iosxe_conn.acquire_priv(desired_priv="exec")

    if auth_secondary[1] is False:
        sync_cisco_iosxe_conn.channel.timeout_ops = 1
        with pytest.warns(UserWarning), pytest.raises(ScrapliTimeout):
            # makes sure we raise a user warning if auth is required for priv escalation
            # this will also fail because the mock ssh server requires a password
            sync_cisco_iosxe_conn.acquire_priv(desired_priv="privilege_exec")
    else:
        sync_cisco_iosxe_conn.acquire_priv(desired_priv="privilege_exec")



"""scrapli.channel.channel"""
import re
import time
from typing import Any, List, Optional, Tuple

from scrapli.channel.base_channel import ChannelBase
from scrapli.decorators import OperationTimeout
from scrapli.exceptions import ScrapliTimeout
from scrapli.helper import get_prompt_pattern, strip_ansi
from scrapli.transport.transport import Transport


class Channel(ChannelBase):
    def __init__(self, transport: Transport, **kwargs: Any) -> None:
        """
        Channel Object

        Args:
            transport: Scrapli Transport class
            kwargs: keyword arguments to pass to ChannelBase

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        super().__init__(transport, **kwargs)

        # ChannelBase supports union of Transport and AsyncTransport, but as this is the
        # "normal" (sync) Channel, transport will always be async
        self.transport: Transport

    def _read_chunk(self) -> bytes:
        """
        Private method to read chunk from channel

        Args:
            N/A

        Returns:
            bytes: output read from channel

        Raises:
            N/A

        """
        new_output = self.transport.read()
        new_output = new_output.replace(b"\r", b"")
        self.logger.debug(f"Read: {repr(new_output)}")
        return new_output

    def _read_until_input(self, channel_input: bytes, auto_expand: Optional[bool] = None) -> bytes:
        """
        Read until all input has been entered.

        Args:
            channel_input: bytes to write to channel
            auto_expand: bool to indicate if a device auto-expands commands, for example juniper
                devices without `cli complete-on-space` disabled will convert `config` to
                `configuration` after entering a space character after `config`; because scrapli
                reads the channel until each command is entered, the command changing from `config`
                to `configuration` will cause scrapli (by default) to never think the command has
                been entered.

        Returns:
            bytes: output read from channel

        Raises:
            N/A

        """
        output = b""

        if not channel_input:
            self.logger.info(f"Read: {repr(output)}")
            return output

        if auto_expand is None:
            auto_expand = self.comms_auto_expand

        # squish all channel input words together and cast to lower to make comparison easier
        processed_channel_input = b"".join(channel_input.lower().split())

        while True:
            output += self._read_chunk()

            # replace any backspace chars (particular problem w/ junos), and remove any added spaces
            # this is just for comparison of the inputs to what was read from channel
            if not auto_expand and processed_channel_input in b"".join(
                output.lower().replace(b"\x08", b"").split()
            ):
                break
            if auto_expand and self._process_auto_expand(
                output=output, channel_input=channel_input
            ):
                break

        self.logger.info(f"Read: {repr(output)}")
        return output

    def _read_until_prompt(self, output: bytes = b"", prompt: str = "") -> bytes:
        """
        Read until expected prompt is seen.

        Args:
            output: bytes from previous reads if needed
            prompt: prompt to look for if not looking for base prompt (self.comms_prompt_pattern)

        Returns:
            bytes: output read from channel

        Raises:
            N/A

        """
        prompt_pattern = get_prompt_pattern(prompt=prompt, class_prompt=self.comms_prompt_pattern)

        while True:
            output += self._read_chunk()
            if self.comms_ansi:
                output = strip_ansi(output=output)
            channel_match = re.search(pattern=prompt_pattern, string=output)
            if channel_match:
                self.logger.info(f"Read: {repr(output)}")
                return output

    def _read_until_prompt_or_time(
        self,
        output: bytes = b"",
        channel_outputs: Optional[List[bytes]] = None,
        read_duration: Optional[float] = None,
    ) -> bytes:
        """
        Read until expected prompt is seen, outputs are seen, or for duration, whichever comes first

        As transport reading may block, transport timeout is temporarily set to the read_duration
        and any `ScrapliTimeout` that is raised while reading is ignored.

        Args:
            output: bytes from previous reads if needed
            channel_outputs: List of bytes to search for in channel output, if any are seen, return
                read output
            read_duration: duration to read from channel for

        Returns:
            bytes: output read from channel

        Raises:
            N/A

        """
        prompt_pattern = get_prompt_pattern(prompt="", class_prompt=self.comms_prompt_pattern)

        if channel_outputs is None:
            channel_outputs = []
        if read_duration is None:
            read_duration = 2.5

        previous_timeout_transport = self.transport.timeout_transport
        self.transport.timeout_transport = int(read_duration)

        start = time.time()
        while True:
            try:
                output += self._read_chunk()
            except ScrapliTimeout:
                pass

            if self.comms_ansi:
                output = strip_ansi(output=output)

            if (time.time() - start) > read_duration:
                break
            if any([channel_output in output for channel_output in channel_outputs]):
                break
            if re.search(pattern=prompt_pattern, string=output):
                break

        self.transport.timeout_transport = previous_timeout_transport

        self.logger.info(f"Read: {repr(output)}")
        return output

    @OperationTimeout(attribute="timeout_ops", message="Timed out determining prompt on device.")
    def get_prompt(self) -> str:
        """
        Get current channel prompt

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        prompt_pattern = get_prompt_pattern(prompt="", class_prompt=self.comms_prompt_pattern)
        self._send_return()
        output = b""
        while True:
            output += self._read_chunk()
            if self.comms_ansi:
                output = strip_ansi(output=output)
            channel_match = re.search(pattern=prompt_pattern, string=output)
            if channel_match:
                current_prompt = channel_match.group(0)
                return current_prompt.decode().strip()

    @OperationTimeout(attribute="timeout_ops", message="Timed out sending input to device.")
    def send_input(
        self, channel_input: str, strip_prompt: bool = True, eager: bool = False
    ) -> Tuple[bytes, bytes]:
        """
        Primary entry point to send data to devices in shell mode; accept input and returns result

        Args:
            channel_input: string input to send to channel
            strip_prompt: strip prompt or not, defaults to True (yes, strip the prompt)
            eager: eager mode reads and returns the `_read_until_input` value, but does not attempt
                to read to the prompt pattern -- this should not be used manually! (only used by
                `send_configs` with the eager flag set)

        Returns:
            raw_result: output read from the channel with no whitespace trimming/cleaning
            processed_result: output read from the channel that has been cleaned up

        Raises:
            N/A

        """
        self._pre_send_input(channel_input=channel_input)

        bytes_channel_input = channel_input.encode()

        with self.transport.session_lock:
            self.logger.info(
                f"Attempting to send input: {channel_input}; strip_prompt: {strip_prompt}"
            )
            self.transport.write(channel_input=channel_input)
            self.logger.debug(f"Write: {repr(channel_input)}")
            output = self._read_until_input(channel_input=bytes_channel_input)
            self._send_return()
            if not eager:
                output = self._read_until_prompt()
        processed_output = self._restructure_output(output=output, strip_prompt=strip_prompt)

        return output, processed_output

    @OperationTimeout(attribute="timeout_ops", message="Timed out sending and reading to device.")
    def send_input_and_read(
        self,
        channel_input: str,
        strip_prompt: bool = True,
        expected_outputs: Optional[List[str]] = None,
        read_duration: Optional[float] = None,
    ) -> Tuple[bytes, bytes]:
        """
        Send a command and read until expected prompt is seen, outputs are seen, or for duration

        Args:
            channel_input: string input to send to channel
            strip_prompt: strip prompt or not, defaults to True (yes, strip the prompt)
            expected_outputs: list of strings to look for in output; if any of these are seen,
                return output read up till that read
            read_duration: float duration to read for

        Returns:
            raw_result: output read from the channel with no whitespace trimming/cleaning
            processed_result: output read from the channel that has been cleaned up

        Raises:
            N/A

        """
        self._pre_send_input(channel_input=channel_input)

        bytes_channel_input = channel_input.encode()
        bytes_channel_outputs = [
            channel_output.encode() for channel_output in expected_outputs or []
        ]

        with self.transport.session_lock:
            self.transport.write(channel_input=channel_input)
            self.logger.debug(f"Write: {repr(channel_input)}")
            self._read_until_input(channel_input=bytes_channel_input)
            self._send_return()
            output = self._read_until_prompt_or_time(
                channel_outputs=bytes_channel_outputs, read_duration=read_duration
            )
        processed_output = self._restructure_output(output=output, strip_prompt=strip_prompt)
        return output, processed_output

    @OperationTimeout(
        attribute="timeout_ops", message="Timed out sending interactive input to device."
    )
    def send_inputs_interact(
        self, interact_events: List[Tuple[str, str, Optional[bool]]]
    ) -> Tuple[bytes, bytes]:
        """
        Interact with a device with changing prompts per input.

        Used to interact with devices where prompts change per input, and where inputs may be hidden
        such as in the case of a password input. This can be used to respond to challenges from
        devices such as the confirmation for the command "clear logging" on IOSXE devices for
        example. You may have as many elements in the "interact_events" list as needed, and each
        element of that list should be a tuple of two or three elements. The first element is always
        the input to send as a string, the second should be the expected response as a string, and
        the optional third a bool for whether or not the input is "hidden" (i.e. password input)

        An example where we need this sort of capability:

        ```
        3560CX#copy flash: scp:
        Source filename []? test1.txt
        Address or name of remote host []? 172.31.254.100
        Destination username [carl]?
        Writing test1.txt
        Password:

        Password:
         Sink: C0644 639 test1.txt
        !
        639 bytes copied in 12.066 secs (53 bytes/sec)
        3560CX#
        ```

        To accomplish this we can use the following:

        ```
        interact = conn.channel.send_inputs_interact(
            [
                ("copy flash: scp:", "Source filename []?", False),
                ("test1.txt", "Address or name of remote host []?", False),
                ("172.31.254.100", "Destination username [carl]?", False),
                ("carl", "Password:", False),
                ("super_secure_password", prompt, True),
            ]
        )
        ```

        If we needed to deal with more prompts we could simply continue adding tuples to the list of
        interact "events".

        Args:
            interact_events: list of tuples containing the "interactions" with the device
                each list element must have an input and an expected response, and may have an
                optional bool for the third and final element -- the optional bool specifies if the
                input that is sent to the device is "hidden" (ex: password), if the hidden param is
                not provided it is assumed the input is "normal" (not hidden)

        Returns:
            raw_result: output read from the channel with no whitespace trimming/cleaning
            processed_result: output read from the channel that has been cleaned up

        Raises:
            N/A

        """
        self._pre_send_inputs_interact(interact_events=interact_events)

        output = b""
        with self.transport.session_lock:
            for interact_event in interact_events:
                channel_input = interact_event[0]
                bytes_channel_input = channel_input.encode()
                channel_response = interact_event[1]
                try:
                    hidden_input = interact_event[2]
                except IndexError:
                    hidden_input = False
                self.logger.info(
                    f"Attempting to send input interact: {channel_input}; "
                    f"\texpecting: {channel_response};"
                    f"\thidden_input: {hidden_input}"
                )
                self.transport.write(channel_input=channel_input)
                self.logger.debug(f"Write: {repr(channel_input)}")
                if not channel_response or hidden_input is True:
                    self._send_return()
                else:
                    output += self._read_until_input(channel_input=bytes_channel_input)
                    self._send_return()
                output += self._read_until_prompt(prompt=channel_response)

        return self._post_send_inputs_interact(output=output)



import sys

from napalm import get_network_driver
from test_data.devices import DEVICES

NAPALM_DEVICE_TYPE_MAP = {
    "cisco_iosxe": "ios",
    "cisco_nxos": "nxos",
    "cisco_iosxr": "iosxr",
    "arista_eos": "eos",
    "juniper_junos": "junos",
}


def prepare_device(test_devices):
    # push base config via napalm to ensure consistent testing experience
    for device in test_devices:
        base_config = DEVICES[device]["base_config"]

        napalm_device_type = NAPALM_DEVICE_TYPE_MAP.get(device)
        napalm_driver = get_network_driver(napalm_device_type)
        napalm_args = {
            "hostname": DEVICES[device]["host"],
            "username": DEVICES[device]["auth_username"],
            "password": DEVICES[device]["auth_password"],
        }
        if device == "arista_eos":
            napalm_args["optional_args"] = {}
            napalm_args["optional_args"]["enable_password"] = DEVICES[device]["auth_password"]
        napalm_conn = napalm_driver(**napalm_args)
        napalm_conn.open()
        napalm_conn.load_replace_candidate(filename=base_config)
        napalm_conn.commit_config()
        napalm_conn.close()


def main():
    test_devices = sys.argv[1].split(",")
    prepare_device(test_devices)


if __name__ == "__main__":
    main()
